/**
 * @file main.cpp
 * @brief Epipolar quasi-Euclidean rectification of images
 * @author Pascal Monasse
 * 
 * Copyright (c) 2014 Pascal Monasse
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "libIO/io_png.h"
#include "sift/demo_lib_sift.h"
#include "libOrsa/fundamental_model.hpp"
#include "rectify/rectify.h"
#include "libTransform/map_image.h"

#include <algorithm>
#include <iostream>
#include <cstdlib>
#include <cfloat>

typedef LWImage<float> FImage;

static const int ITER_RANSAC=10000;
static const int SPLINE_ORDER=5;

/// Load float image from PNG file.
static bool load(const char* nameFile, FImage& im)
{
    size_t sw,sh,sc;
    im.data = io_png_read_f32(nameFile, &sw,&sh,&sc);
    if(! im.data) {
        std::cerr << "Unable to load image file " << nameFile << std::endl;
        return false;
    }

    im.w = static_cast<int>(sw);
    im.h = static_cast<int>(sh);
    im.comps = static_cast<int>(sc);
    return true;
}

/// Save float image into PNG file.
static bool save(const char* nameFile, FImage& im) {
    // Put in white invalid pixels
    for(int i=im.comps*im.w*im.h-1; i>=0; i--)
        if(! is_number(im.data[i]))
            im.data[i] = 255.0f;
    if(io_png_write_f32(nameFile, im.data, im.w, im.h, im.comps) != 0) {
        std::cerr << "Error writing file " << nameFile << std::endl;
        return false;
    }
    return true;
}

/// Lexicographical ordering of matches. Used to remove duplicates.
static bool operator<(const Match& m1, const Match& m2)
{
    if(m1.x1 < m2.x1) return true;
    if(m1.x1 > m2.x1) return false;

    if(m1.y1 < m2.y1) return true;
    if(m1.y1 > m2.y1) return false;

    if(m1.x2 < m2.x2) return true;
    if(m1.x2 > m2.x2) return false;

    return (m1.y2 < m2.y2);
}

static bool operator==(const Match& m1, const Match& m2)
{
    return (m1.x1==m2.x1 && m1.y1==m2.y1 &&
            m1.x2==m2.x2 && m1.y2==m2.y2);
}

/// Convert to gray
float* gray(const FImage& im) {
    const float DIV = 1.0f/im.comps;
    const int sz = im.w*im.h;
    float* data = new float[sz];
    for(int i=sz-1; i>=0; i--) {
        float v = im.data[i];
        for(int j=1; j<im.comps; j++)
            v += im.data[i+j*sz];
        data[i] = v*DIV;
    }
    return data;
}

/// Find SIFT point correspondences between both images.
static void sift_correspondences(const FImage& im1, const FImage& im2,
                                 matchingslist& m) {
    siftPar param;
    default_sift_parameters(param);
    param.DoubleImSize=0;

    float* data;

    data = im1.data;
    if(im1.comps!=1) data = gray(im1);
    keypointslist keyp1, keyp2;
    compute_sift_keypoints(data, keyp1, im1.w, im1.h, param);
    std::cout<< "sift: im1: " << keyp1.size() <<std::flush;
    if(im1.comps!=1) delete [] data;

    data = im2.data;
    if(im2.comps!=1) data = gray(im2);
    compute_sift_keypoints(data, keyp2, im2.w, im2.h, param);
    std::cout<<      " im2: " << keyp2.size() <<std::flush;
    if(im2.comps!=1) delete [] data;

    matchingslist matchings;
    compute_sift_matches(keyp1, keyp2, m, param);
    std::cout<< " matches: " << m.size() <<std::endl;
}

/// Select valid point correspondences wrt a fundamental matrix by AC-RANSAC.
static bool RANSAC(std::vector<Match>& m, int w, int h) {
    const int n = static_cast<int>(m.size());
    if(n<8) {
        std::cerr << "AC-RANSAC needs at least 8 correspondences" << std::endl;
        return false;
    }
    libNumerics::matrix<double> pts1(2,n), pts2(2,n);

    for(int i=0; i<n; ++i) {
        pts1(0,i) = m[i].x1;
        pts1(1,i) = m[i].y1;
        pts2(0,i) = m[i].x2;
        pts2(1,i) = m[i].y2;
    }
    orsa::FundamentalModel model(pts1,w,h, pts2,w,h);
    libNumerics::matrix<double> F(3,3);
    std::vector<int> inliers;
    if(model.orsa(inliers,ITER_RANSAC,0,&F,true) > 0.0)
        return false;
    std::cout << "F= " << F <<std::endl; // Denormalization

    std::sort(inliers.begin(), inliers.end());
    for(size_t i=0; i<inliers.size(); i++)
        m[i] = m[inliers[i]];
    m.erase(m.begin()+inliers.size(), m.end());
    return true;
}

/// Compute and print min and max disparity
void printDisparity(const std::vector<Match>& match,
                    const libNumerics::Homography& Hl,
                    const libNumerics::Homography& Hr) {
    std::vector<Match>::const_iterator it=match.begin();
    double min=DBL_MAX, max=-DBL_MAX;
    for(; it != match.end(); ++it) {
        double xl=it->x1, yl=it->y1;
        Hl(xl,yl);
        double xr=it->x2, yr=it->y2;
        Hr(xr,yr);
        xr -= xl;
        if(xr < min)
            min = xr;
        if(xr > max)
            max = xr;
    }
    std::cout << "Disparity: "
              << (int)floor(min) << " " << (int)ceil(max) << std::endl;
}

/// Rectification of two images
int main(int argc, char* argv[]) {
    // (0) Input
    if(argc != 5) {
        std::cerr << "Usage: " << argv[0]
                  << " im1.png im2.png"
                  << " out1.png out2.png" << std::endl;
        return 1;
    }

    FImage im1, im2;
    if(! load(argv[1], im1) ||
       ! load(argv[2], im2))
        return 1;

    // (1) SIFT
    std::vector<Match> m;
    sift_correspondences(im1, im2, m);
    std::sort(m.begin(), m.end());
    std::vector<Match>::iterator end = std::unique(m.begin(), m.end());
    if(end != m.end()) {
        std::cout << "Remove " << std::distance(end,m.end())
                  << "/" << m.size() << " duplicate matches"<<std::endl;
        m.erase(end, m.end());
    }
    
    // (2) AC-RANSAC
    long int seed = (long int)time(NULL);
    srand(seed);
    if(! RANSAC(m, im1.w, im1.h)) {
        std::cerr << "Error in RANSAC algorithm, aborting" << std::endl;
        return 1;
    }

    // (3) Compute rectifying homographies
    if(im1.w!=im2.w || im1.h!=im2.h) {
        std::cerr << "Images must have same size for this rectification method"
                  << std::endl;
        return 1;
    }
    libNumerics::Homography H1, H2;
    std::pair<float,float> e = compRectif(im1.w, im1.h, m, H1, H2);
    std::cout << "Initial rectification error: " <<e.first <<" pix" <<std::endl;
    std::cout << "Final rectification error: " << e.second <<" pix" <<std::endl;
    printDisparity(m, H1, H2);

    // (4) Apply homographies
    FImage out1 = alloc_image<float>(im1.w, im1.h, im1.comps);
    map_image(im1, H1, out1, SPLINE_ORDER, false, false);
    FImage out2 = alloc_image<float>(im2.w, im2.h, im2.comps);
    map_image(im2, H2, out2, SPLINE_ORDER, false, false);

    // (5) Output images
    if(! save(argv[3], out1) ||
       ! save(argv[4], out2))
        return 1;

    free(im1.data);
    free(im2.data);
    free(out1.data);
    free(out2.data);
    return 0;
}
